(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Subterm_Criterion_Impl
imports
  Subterm_Criterion
  Trs_Impl
begin

datatype 'f projL = Projection "('f \<times> nat) list"

definition
  check_strict ::
    "('f::show, 'v::show) rules \<Rightarrow>
     (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option) \<Rightarrow>
     'f proj \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows check"
where
  "check_strict R rseqm p r = (
     let s = proj_term p (fst r); t = proj_term p (snd r) in
     case rseqm r of
       None \<Rightarrow> check_supt s t
     | Some rseq \<Rightarrow> (if length rseq = 0
       then check_supt s t
       else (
         let u = snd (snd (last rseq)) in (
           check_rsteps R rseq s u
         \<guillemotright> check_supteq u t
         )
       )
     )
   )"

lemma check_strict_sound[simp]:
  assumes "isOK (check_strict R rseqm p r)"
  shows "(proj_term p (fst r), proj_term p (snd r)) \<in> (op \<rhd> \<union> rstep (set R))^+"
    (is "(?s, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+")
proof (cases "rseqm r")
  case None
  from assms[unfolded check_strict_def None]
    have "isOK (check_supt ?s ?t)" by simp
  hence "(?s, ?t) \<in> op \<rhd>" by simp
  thus ?thesis by auto
next
  case (Some rseq)
  show ?thesis
  proof (cases "length rseq = 0")
    case True
    with assms[unfolded check_strict_def Some]
      have "isOK (check_supt ?s ?t)" by simp
    hence "(?s, ?t) \<in> op \<rhd>" by simp
    thus ?thesis by auto
  next
    let ?u = "snd (snd (last rseq))"
    case False
    with assms[unfolded check_strict_def Some]
      have "isOK (check_rsteps R rseq ?s ?u \<guillemotright> check_supteq ?u ?t)" unfolding Let_def by simp
    hence "(?s, ?u) \<in> (rstep (set R))^^(length rseq)" and "(?u, ?t) \<in> op \<unrhd>" by simp_all
    with False obtain n where "(?s, ?u) \<in> (rstep (set R))^^(Suc n)" by (induct rseq) simp_all
    hence "(?s, ?u) \<in> (rstep (set R))^+" using trancl_power[of _ "rstep (set R)"] by blast
    from `(?u, ?t) \<in> op \<unrhd>` show ?thesis unfolding supteq_supt_conv
    proof
      assume "?u = ?t"
      from `(?s, ?u) \<in> (rstep (set R))^+`
        show ?thesis unfolding `?u = ?t` using trancl_union_right[of "rstep (set R)" "op \<rhd>"] by blast
    next
      assume "(?u, ?t) \<in> op \<rhd>"
      hence "(?u, ?t) \<in> (op \<rhd> \<union> rstep (set R))^+" by auto
      moreover from `(?s, ?u) \<in> (rstep (set R))^+`
        have "(?s, ?u) \<in> (op \<rhd> \<union> rstep (set R))^+" using trancl_union_right by blast
      ultimately show ?thesis by simp
    qed
  qed
qed

definition
  check_weak :: "'f proj \<Rightarrow> ('f::show, 'v::show) rule \<Rightarrow> shows check"
where
  "check_weak p r \<equiv> check (proj_term p (fst r) = proj_term p (snd r))
     (''the projected lhs is not equal to the projected rhs'' +#+ shows_nl)
   <+? (\<lambda>s. ''Could not orient rule '' +#+ shows_rule r +@+ '', since'' +#+ shows_nl +@+
          shows (proj_term p (fst r)) +@+ '' != '' +#+ shows (proj_term p (snd r)) +@+
          shows_nl +@+ s)"

lemma check_weak_sound[simp]:
  assumes "isOK (check_weak p r)" shows "proj_term p (fst r) = proj_term p (snd r)"
using assms unfolding check_weak_def by auto

fun
  check_subterm_criterion_proc ::
    "'f proj \<Rightarrow>
     (('f::show, 'v::show) rule \<Rightarrow> ('f::show, 'v::show) rseq option) \<Rightarrow>
     ('f, 'v) rules \<Rightarrow>
     ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v) rule_map \<Rightarrow> shows check"
where
  "check_subterm_criterion_proc p rseqm P P' R None = OK"
| "check_subterm_criterion_proc p rseqm P P' R (Some rm) = do {
     check_allm (\<lambda>(l, r). check_no_var l \<guillemotright> check_no_var r \<guillemotright> check_no_defined_root rm r) P;
     check_allm (check_strict R rseqm p) (list_diff P P');
     check_allm (check_weak p) P'
   }"

lemma check_subterm_criterion_proc_sound[simp] :
  "ci_only_pairs_proc_impl (\<lambda>P R rm rrm P'. isOK (check_subterm_criterion_proc p rseqm P P' R rm))"
unfolding ci_only_pairs_proc_impl_def proof (intro allI impI)
  fix P R P' s t \<sigma> assume ok: "isOK (check_subterm_criterion_proc p rseqm P P' R (computeRuleMap R))"
    and mchain: "min_ichain (set P, set R) s t \<sigma>"
  hence "SN_elt (rstep (set R)) (t 0 \<cdot> \<sigma> 0)" by (simp add: minimal_cond_def)
  then obtain rm where rm: "computeRuleMap R = Some rm" using computeRuleMap_None_not_SN by (cases "computeRuleMap R", auto)
  let ?\<pi> = "\<lambda>t. proj_term p t"
  from ok[unfolded rm] 
    have cond: "\<forall>(l, r)\<in>set P. \<not> is_var l \<and> \<not> is_var r \<and> \<not> defined (set R) (the (root r), num_args r)" by (simp add: split_def check_no_defined_root_sound[OF rm])
  moreover have strict: "\<forall>(l, r)\<in>set P - set P'. (?\<pi> l, ?\<pi> r) \<in> (op \<rhd> \<union> rstep (set R))^+"
  proof -
    from ok[unfolded rm]
      have "isOK (check_allm (check_strict R rseqm p) (list_diff P P'))" by simp
      thus ?thesis unfolding split_def using check_strict_sound[of R rseqm p] by (simp, blast)
  qed
  moreover from ok[unfolded rm]
    have weak: "\<forall>(l, r)\<in>set P'. ?\<pi> l = ?\<pi> r" by (simp add: split_def)
  ultimately
    show "\<exists>i. ichain (set P', set R) (shift s i) (shift t i) (shift \<sigma> i)"
    using ci_subterm_criterion_proc[of p, unfolded ci_proc_def subterm_criterion_proc_def subterm_criterion_cond_def,
      THEN spec[of _ "set P"], THEN spec[of _ "set R"], THEN spec[of _ "set P'"], THEN spec[of _ "set R"]]
    using mchain by blast
qed

fun
  create_proj :: "'f projL \<Rightarrow> 'f proj"
where
  "create_proj (Projection p) f = (case lookup f p of None \<Rightarrow> 0 | Some n \<Rightarrow> n)"

fun
  create_rseq_map :: "(('f, 'v) rule \<times> ('f, 'v) rseq) list \<Rightarrow> (('f, 'v) rule \<Rightarrow> ('f, 'v) rseq option)"
where
  "create_rseq_map rseqs r = lookup r rseqs"

fun
  create_subterm_criterion_proc ::
    "'f projL \<Rightarrow> (('f, 'v) rule \<times> ('f, 'v) rseq) list \<Rightarrow>
     ('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v) rule_map
     \<Rightarrow> shows check"
where
  "create_subterm_criterion_proc p rseqs P rP R rm = (let \<pi> = create_proj p; rseqm = create_rseq_map rseqs in
     check_subterm_criterion_proc \<pi> rseqm P rP R rm)"

lemma subterm_criterion_proc_ci:
  "ci_only_pairs_proc_impl (\<lambda>P R rm rrm rP. isOK (create_subterm_criterion_proc p rseqs P rP R rm))"
proof -
  let ?p = "create_proj p"
  let ?rseqs = "create_rseq_map rseqs"
  have simpme: "create_subterm_criterion_proc p rseqs = check_subterm_criterion_proc ?p ?rseqs"
  by (rule ext, rule ext, rule ext, rule ext, simp add: Let_def)
  show ?thesis by (simp only: simpme, rule check_subterm_criterion_proc_sound)
qed

declare create_subterm_criterion_proc.simps[simp del]

end
